function [WJt, VJt, att] = cmodwt(X, g, h, nlevels, boundary, retainVJ)

%   cmodwt -- Compute the complex (partial) maximal overlap discrete wavelet transform (CMODWT).
%             Modification of MODWT algorithm accepting complex wavelet filter
%
% NAME
%   cmodwt -- Compute the complex (partial) maximal overlap discrete wavelet transform (CMODWT).
%
% SYNOPSIS
%   [WJt, VJt, att] = cmodwt(X, g, h, nlevels, boundary, retainVJ)
%
% INPUTS
%   * X          -- set of observations (vector of length NX or matrix of size NX x Nchan)
%
%   * g          -- complex wavelet transform scaling filter (optional)
%                   default: LinaMayrand 3.1
% 
%   * h          -- complex wavelet transform wavelet filter (optional)
%                   default: LinaMayrand 3.1
%
%   * nlevels    -- maximum level J0 (integer) (optional)
%                   default: J0 = floor(log2( (N / (L - 1)) - 1)),
%                   computed by 'conservative' procedure in MODWT
%                   N - length of input, L - length of filter
%                   
%   * boundary   -- boundary conditions to use (string) (optional)
%                   Valid values: 'circular'/'periodic' or 'reflection'
%                   Default: 'periodic'
%   * reatinVJ   -- Boolean flag indicating whether to scaling coefficients
%                   have been retained at all levels. (optional)
%                   Values:  1 = true,  all VJ retained, 
%                            0 = false, VJ retained for J0 level.
%                   Default: 0, VJ retained only at J0 level.
%
% OUTPUTS
%   * WJt        -- MODWT wavelet coefficents (NW x J x NChan array).
%   * VJt        -- MODWT scaling coefficients (NW x {1,J} x NChan vector).
%   * att        -- MODWT transform attributes (struct).
%
%   The output argument att is a structure with the following fields:
%   * g          -- complex wavelet transform scaling filter
%   * h          -- complex wavelet transform wavelet filter
%   * NX         -- number of observations in original series (= length(X))
%   * NW         -- number of wavelet coefficients
%   * J0         -- number of levels of partial decompsition.
%   * NChan      -- number of channels in a multivariate dataset.
%   * Boundary   -- boundary conditions applied.
%   * retainVJ   -- Boolean flag indicating whether VJ scaling coefficients
%                   at all levels have been retained (1= true) or not (0 =
%                   false).
%
% DESCRIPTION
%   cmodwt calculates the wavelet and scaling coefficients using the complex maximal
%   overlap discrete wavelet transform (CMODWT).
%
% CREATED BY
%   Norbert Remenyi
%
% CREATION DATE
%   2012-05-10
%
% CREDITS
%   Based on the original function (modwt.m) by Charles R. Cornish.

% REFERENCES
%   Percival, D. B. and A. T. Walden (2000) Wavelet Methods for
%   Time Series Analysis. Cambridge: Cambridge University Press. 
%


%% If X is a vector, make X a column vector
if isvector(X)
  X = X(:);
end

%%  NX    = length of original series
%%  NChan = number of channels for multi-variant dataset.
[NX, NChan] = size(X);

%% Check number of arguments
if nargin==5
    retainVJ = [];
end;
if nargin==4
    retainVJ = [];
    boundary = [];
end;
if nargin==3
    nlevels = [];
    retainVJ = [];
    boundary = [];
end;
if nargin==2
    h = [];
    nlevels = [];
    retainVJ = [];
    boundary = [];
end;
if nargin==1
    g = [];
    h = [];
    nlevels = [];
    retainVJ = [];
    boundary = [];
end;

%% Use LinaMayrand 3.1 filter as default if filters are missing
if (isempty(g) && isempty(h))
    g = [-0.066291260736239 - 0.085581649610182i, 0.110485434560398 - 0.085581649610182i,...
        0.662912607362388 + 0.171163299220364i, 0.662912607362388 + 0.171163299220364i,...
        0.110485434560398 - 0.085581649610182i, -0.066291260736239 - 0.085581649610182i];
    h = [-0.066291260736239 + 0.085581649610182i, -0.110485434560398 - 0.085581649610182i,...
        0.662912607362388 - 0.171163299220364i, -0.662912607362388 + 0.171163299220364i,...
        0.110485434560398 + 0.085581649610182i, 0.066291260736239 - 0.085581649610182i];
end;
% 'MakeCONFilter.m' in Wavelab can generate coefficients of Daubechies maxflat
% symmetric complex orthonormal wavelet filters    

%% If only scaling wavelet filter g is defined, compute wavelet filter
if (isempty(h))
    L = length(g);
    for i=0:L-1
        h(i+1) = (-1)^i*conj(g(L-i));
    end;
end;

%% Get tilda filters for MODWT, dividing by sqrt(2)
gt = g/sqrt(2);
ht = h/sqrt(2);

%% Get real and imaginary parts of filters
gtR = real(gt);
gtI = imag(gt);
htR = real(ht);
htI = imag(ht);

%% Length of filter
L = length(g);

%%  If nlevels is an integer > 0, set J0 = nlevels.
%%  If nlevels missing, select J0 based on 'conservative' method explained in MODWT.
if (isempty(nlevels)) 
  J0 = floor(log2( (NX / (L - 1)) - 1));
elseif (nlevels > 0)
    J0 = nlevels;
  else
    error('nlevels must be an integer greater than 0.');
end

if (J0 < 0)
  error('J0 must be greater than 0.');
end

if (2^J0 > NX)
  warning('JO > log2(Number of samples).');
end

%% If retainVJ is not specified, use 0 as default
if (isempty(retainVJ))
    retainVJ = 0;
end;

%% If boundary condition is not specified, use 'reflection' as default
if (isempty(boundary))
    boundary = 'periodic';
end;

%% Initialize the scale (Vin) for first level by setting it equal to X
%% using specified  boundary conditions
switch boundary
  case 'reflection'
   Xin = cat(1, X, flipdim(X, 1));
  case {'circular', 'periodic'}
   Xin = X;
  otherwise
   error('Invalid boundary method.');
end

%% NW = length of the extended series = number of coefficients
NW = size(Xin, 1);


%% Pre-allocate memory.
WJt = NaN([NW, J0, NChan], 'double');
if (retainVJ)
  VJt = NaN([NW, J0, NChan], 'double');
else
  VJt = NaN([NW, 1, NChan], 'double');
end

%% Do the MODWT.
for i = 1:NChan
  Vin = Xin(:,i);
  VinR = real(Vin);
  VinI = imag(Vin);
  for (j = 1:J0)
    [Wt_jR Wt_jI VoutR VoutI] = cmodwtj(VinR, VinI, htR, htI, gtR, gtI, j);
    WJt(:,j,i) = complex(Wt_jR,Wt_jI);
    VinR = VoutR;
    VinI = VoutI;
    if (retainVJ)
      VJt(:,j,i) = complex(VoutR,VoutI);
    end
  end
  if (~retainVJ)
    VJt(:,1,i) = complex(VoutR,VoutI);
  end
end

%% Update attributes
att.g          = g;
att.h          = h;
att.NX         = NX;
att.NW         = NW;
att.J0         = J0;
att.NChan      = NChan;
att.Boundary   = boundary;
att.RetainVJ   = retainVJ;


return
