% This function performs the 2D complex maximum overlap scale-mixing 
% wavelet tranform on an input image, estimates its wavelet coefficients 
% by an empirical Bayes procedure, and computes the inverse transform 
%
% Input
%  Ln       -- noisy input image
%  sigma    -- known noise standard deviation
%  gc       -- complex wavelet filter coefficients
%  coarsest -- coarsest level of wavelet decomposition
%
% Output
%  L_tr     -- denoised image
%
%  The program calls to:
%       Wavmat.m    -- Transformation matrix of fast wavelet transform 
%                       (periodic orthogonal)
%       loglik_2D.m -- Evaluates the log-likelihood function for 
%                       wavelet coefficients in a specific subband
%       cmodwt.m    -- Computes the complex (partial) maximal overlap
%                       discrete wavelet transform (CMODWT).
% -------------------------------------------------------------------------

function L_tr = rec_cMOSM_EB(Ln,sigma,gc,coarsest)

n = length(Ln); %size of noisy image (squared)
finest = log2(n)-1; %finest level of wavelet decomposition

% Compute wavelet transformation matrix
Wmod = [];
In = eye(n);
J = finest-coarsest+1; % number of levels of detail
Lf = length(gc);
for i=0:Lf-1
    hc(i+1) = (-1)^i*conj(gc(Lf-i));
end; 
for i=1:n
[W V] = cmodwt(In(:,i), gc, hc, J);    
Wmod(:,i) = [V;fliplr(W(:))];
clearvars V W;
end;
clearvars In;

% 2D complex maximum overlap scale-mixing wavelet transform
D1 = Wmod*Ln*Wmod';
clearvars Ln;

%%%%%%%%% Covariance structure of noise in wavelet domain %%%%%%%
Wmodt = conj(Wmod');
DD1 = diag(conj(Wmod)*conj(Wmodt))*conj(diag(Wmod*Wmodt))';
DD1 = DD1(:);
DD2 = diag(Wmod*Wmod')*conj(diag(conj(Wmod)*Wmodt))';
DD2 = DD2(:);
clearvars Wmodt;
Cov_ReRed = sigma^2/2*(real(DD2)+real(DD1));
Cov_ImImd = sigma^2/2*(real(DD2)-real(DD1));
Cov_ReImd = -sigma^2/2*(imag(DD1)+imag(DD2));
Cov_ReRe = reshape(Cov_ReRed,n*(J+1),n*(J+1));
Cov_ImIm = reshape(Cov_ImImd,n*(J+1),n*(J+1));
Cov_ReIm = reshape(Cov_ReImd,n*(J+1),n*(J+1));
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clearvars DD1 DD2;

%%%%%%%%%%%%%% Wavelet shrinkage by empirical Bayes procedure %%%%%%%%%%%%%
% i,j represent subband position in the tesselation of the 2D transform
% rowix, colix are the indices of wavelet coefficients in subband (i,j)

for i = 1:J+1
    i
    for j = 1:J+1
        if (i~=1 || j~=1) % Keep coefficients on coarsest level
        rowix = (i-1)*n+1:i*n;
        colix = (j-1)*n+1:j*n;
        
        % All complex wavelet coefficients in subband (i,j)
        d0 = D1(rowix,colix);
        d0 = d0(:);            
        d =[real(d0) imag(d0)];
        
        % Covariance structure of a complex wavelet coefficient in 
        % subband (i,j)
        S = [Cov_ReRe(rowix(1),colix(1)) Cov_ReIm(rowix(1),colix(1));...
            Cov_ReIm(rowix(1),colix(1)) Cov_ImIm(rowix(1),colix(1))];
                            
        %%%% Empirical Bayes procedure to estimate hyperparameters %%%%
       
        % Initial estimate for covariance matrix of wavelet coefficients
        D = cov(d);
        % Starting values (rescaled) for mixing weight and covariance matrix
        x0 = [0.5 D(1,1) D(1,2)/sqrt(D(1,1)*D(2,2)) D(2,2)];
        % Upper and lower bounds
        tol = 0.001;
        lb = [tol tol tol-1 tol];
        ub = [1-tol Inf 1-tol Inf];
        % Interior point method to minimize negative of log-likelihood
        % and to find hyperparameters
        options = optimset('Algorithm','interior-point','Display','off');
        par_opt = fmincon(@(par) loglik_2D(par,d,S),x0,[],[],[],[],...
                                                lb,ub,[],options); 
        
        % Estimated mixing weight and covariance matrix (rescale back)      
        p = par_opt(1);
        V = [par_opt(2),par_opt(3)*sqrt(par_opt(2)*par_opt(4));...
                par_opt(3)*sqrt(par_opt(2)*par_opt(4)),par_opt(4)];
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        % Posterior mean shrinkage rule
        detVpS = det(V+S);
        invVpS = inv(V+S);
        detS = det(S);
        invS = inv(S);
        invinvVinvS = inv(inv(V)+inv(S));
        
        for r=rowix
            for s=colix              
                dd = [real(D1(r,s)) imag(D1(r,s))]';    
                p_hat = p*1/(2*pi*sqrt( detVpS )).*exp(-0.5*dd'*invVpS*dd) / ...
                (p*1/(2*pi*sqrt( det(V+S) )).*exp(-0.5*dd'*invVpS*dd) +...
                (1-p)/(2*pi*sqrt( detS )).*exp(-0.5*dd'*invS*dd) );
                dd_tr = p_hat * ( invinvVinvS *invS*dd);
                D1(r,s) = dd_tr(1) + dd_tr(2)*sqrt(-1);
                
            end;
        end;
        end;
        
    end;
end;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Perform inverse wavelet transform and take real part
L_tr = real(Wmod' * D1 * Wmod);

clearvars Wmod D1 Cov_ReRed Cov_ImImd Cov_ReImd Cov_ReRe Cov_ImIm Cov_ReIm;

