% This function performs the 2D scale-mixing complex wavelet tranform 
% on an input image, estimates its wavelet coefficients by an empirical 
% Bayes procedure, and computes the inverse transform 
%
% Input
%  Ln       -- noisy input image
%  sigma    -- known noise standard deviation
%  wf       -- complex wavelet filter coefficients
%  coarsest -- coarsest level of wavelet decomposition
%
% Output
%  L_tr     -- denoised image
%
%  The program calls to:
%       Wavmat.m    -- Transformation matrix of fast wavelet transform 
%                       (periodic orthogonal)
%       loglik_2D.m -- Evaluates the log-likelihood function for 
%                       wavelet coefficients in a specific subband
% -------------------------------------------------------------------------

function L_tr = rec_cSM_EB(Ln,sigma,wf,coarsest)

n = length(Ln); %size of noisy image (squared)
finest = log2(n)-1; %finest level of wavelet decomposition

% Wavelet transformation matrix
W1 = Wavmat(wf, n, log2(n)-coarsest,  0);

% 2D scale-mixing complex wavelet transform
D1 = W1 * Ln * W1';

% Keep coefficients on coarsest level
D1_coarse = D1(1:2^coarsest,1:2^coarsest);

% If noise level is assumed to be unknown, it can be estimated from 
% the coefficients at the finest level of detail
% niz_re = real(D1(dyad(finest),dyad(finest)));
% niz_im = imag(D1(dyad(finest),dyad(finest)));
% sigmasq = (mad(niz_re(:),1)/0.6745)^2+(mad(niz_im(:),1)/0.6745)^2;
% sigma = sqrt(sigmasq);

%%%%%%%%% Covariance structure of noise in wavelet domain %%%%%%%
W1t = conj(W1');
DD = diag(W1*W1t)*conj(diag(conj(W1)*conj(W1t)))';
clearvars W1t;
DD = DD(:);
Cov_ReRed = sigma^2/2*(ones(n^2,1)+real(DD));
Cov_ImImd = sigma^2/2*(ones(n^2,1)-real(DD));
Cov_ReImd = sigma^2/2*imag(DD);
Cov_ReRe = reshape(Cov_ReRed,n,n);
Cov_ImIm = reshape(Cov_ImImd,n,n);
Cov_ReIm = reshape(Cov_ReImd,n,n);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clearvars DD;

%%%%%%%%%%%%% Wavelet shrinkage by empirical Bayes procedure %%%%%%%%%%%%%
% i,j represent subband position in the tesselation of the 2D transform
% rowix, colix are the indices of wavelet coefficients in subband (i,j)
for i = coarsest:finest+1
    for j = coarsest:finest+1
        if i==coarsest 
            rowix = 1:2^i;
        else rowix = 2^(i-1)+1:2^i;
        end;
        if j==coarsest 
            colix = 1:2^j;
        else colix = 2^(j-1)+1:2^j;
        end;
                        
        % All complex wavelet coefficients in subband (i,j)
        d0 = D1(rowix,colix);
        d0 = d0(:);
        d =[real(d0) imag(d0)];
        
        % Covariance structure of a complex wavelet coefficient in 
        % subband (i,j)
        S = [Cov_ReRe(rowix(1),colix(1)) Cov_ReIm(rowix(1),colix(1));...
            Cov_ReIm(rowix(1),colix(1)) Cov_ImIm(rowix(1),colix(1))];
                            
        %%%% Empirical Bayes procedure to estimate hyperparameters %%%%
        
        % Initial estimate for covariance matrix of wavelet coefficients
        D = cov(d);
        % Starting values (rescaled) for mixing weight and covariance matrix
        x0 = [0.5 D(1,1) D(1,2)/sqrt(D(1,1)*D(2,2)) D(2,2)];
        % Upper and lower bounds
        tol = 0.001;
        lb = [tol tol tol-1 tol];
        ub = [1-tol Inf 1-tol Inf];
        % Interior point method to minimize negative of log-likelihood
        % and to find hyperparameters
        options = optimset('Algorithm','interior-point','Display','off');
        par_opt = fmincon(@(par) loglik_2D(par,d,S),x0,[],[],[],[],...
                                            lb,ub,[],options); 
        
        % Estimated mixing weight and covariance matrix (rescale back)
        p = par_opt(1);
        V = [par_opt(2),par_opt(3)*sqrt(par_opt(2)*par_opt(4));...
            par_opt(3)*sqrt(par_opt(2)*par_opt(4)),par_opt(4)];
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % Posterior mean shrinkage rule
        invVpS = inv(V+S);
        invS = inv(S);
        invinvVinvS = inv(inv(V)+invS);
        
        for r = rowix
            for s = colix              
                dd = [real(D1(r,s)) imag(D1(r,s))]';    
                p_hat = p*1/(2*pi*sqrt( det(V+S) )).*exp(-0.5*dd'*invVpS*dd) / ...
                (p*1/(2*pi*sqrt( det(V+S) )).*exp(-0.5*dd'*invVpS*dd) + ...
                (1-p)/(2*pi*sqrt( det(S) )).*exp(-0.5*dd'*invS*dd) );
                dd_tr = p_hat * ( invinvVinvS*invS*dd);
                D1(r,s) = dd_tr(1) + dd_tr(2)*sqrt(-1);
            end;
        end;
                
    end;
end;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Keep coefficients on coarsest level
D1(1:2^coarsest,1:2^coarsest) = D1_coarse;

% Perform inverse wavelet transform and take real part
L_tr = real(W1' * D1 * W1);

clearvars W1 D1 Cov_ReRed Cov_ImImd Cov_ReImd Cov_ReRe Cov_ImIm Cov_ReIm;
